/*
 * Copyright 1992 the Board of Trustees of the Leland Stanford Junior
 * University. Official permission to use this software is included in
 * the documentation. It authorizes you to use this file for any
 * non-commercial purpose, provided that this copyright notice is not
 * removed and that any modifications made to this file are commented
 * and dated in the style of the example below.
 */

/*
 *
 *  source file:   ./xtpanel/input.c
 *
 *  Dave Nichols (SEP), July 3 1993
 *      Inserted this sample edit history entry.
 *      Please log any further modifications made to this file:
 *	Wrote input.c.
 *  Steve Cole (SEP), October 7 1993
 *      Added pipe and output objects.
 */

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include <stdio.h>
#include <math.h>

#include <sys/ioctl.h>
#include <sys/types.h>

#include <X11/Xos.h>

#ifndef STDIN_FILENO
#define STDIN_FILENO    0
#endif
#ifndef STDOUT_FILENO
#define STDOUT_FILENO   1
#endif

#include "object.h" 
#include "tree.h"
#include "builders.h"
#include "string_buf.h"

extern Widget toplevel;

typedef struct inputinfo {
    char *type;
    unsigned long fd;
    unsigned long ofd;
    XtInputId id;
    int running;
    char *separator;
} _inpinfo;

extern int stdin_used;
extern XtAppContext app_con;

extern void input_callback();
extern void input_sync();

/* get the file descriptor to use for input or output */
int get_fd( stream, objname, type  )
char* stream;
char* objname;
char* type;
{
FILE* infile;
FILE* outfile;

if( !strcmp( stream, "none" )  ){
  return -1;
}

if( !strcmp( type, "input" ) ){
  if( !strcmp( stream, "stdin" )  ){
    /* use stdin if available */
    if( stdin_used ){
        fprintf(stderr,
        "input object %s cannot use stdin, already in use\n",objname);
               exit(-1);
    }
    stdin_used = 1;
    infile = stdin;

  }else{
    /* start the command */
    if( ( infile = popen( stream, "r" )) == 0 ){
       fprintf(stderr, "unable to execute input command %s for object %s\n", 
                 stream,objname);
       exit(-1);
    }
  }
  return ( fileno(infile) );
} else {
  if( !strcmp( stream, "stdout" )  ){
    outfile = stdout;

  }else{
    /* start the command */
    if( ( outfile = popen( stream, "w" )) == 0 ){
       fprintf(stderr, "unable to execute output command %s for object %s\n", 
                 stream,objname);
       exit(-1);
    }
  }
  return ( fileno(outfile) );
}
}

/* for object type pipe we need two pipes for two-way communication */
void get_pipes( stream, objname, inpipe, outpipe  )
char* stream;
char* objname;
int *inpipe, *outpipe;
{
FILE* infile, *outfile;
int pipe1[2], pipe2[2];
int child ;

if( !strcmp( stream, "none" )  ){
  *inpipe = -1;
  *outpipe = -1;
  return;
}

if( !strcmp( stream, "stdin" )  ){
    /* use stdin if available */
    if( stdin_used ){
        fprintf(stderr,
        "pipe object %s cannot use stdin, already in use\n",objname);
               exit(-1);
    }
    stdin_used = 1;
    infile = stdin;
    *inpipe = fileno(infile);
    *outpipe = fileno(stdout);
    return ;

}else{
    /* create the pipes */
    if( pipe(pipe1) < 0 || pipe(pipe2) < 0) {
       fprintf(stderr, "unable to open pipes for object %s\n", 
                objname);
       exit(-1);
    }
 
    /* fork off a child process that will run the desired job */
    if ((child = fork()) == -1){
	perror("fork");
	exit(-1);

    }else if (child) {                       /* this is the parent */
	infile = fdopen(pipe1[0],"r");
	outfile = fdopen(pipe2[1],"w");
        /* close the unused pipe ends */
	if (close(pipe1[1]) != 0)
		fprintf(stderr,"parent: could not close pipe1[1]\n");
	if (close(pipe2[0]) != 0)
		fprintf(stderr,"parent: could not close pipe2[0]\n");
	*inpipe = fileno(infile);
	*outpipe = fileno(outfile);
        return; 

    } else {                                /* this is the child */
	/* reset stdin, stdout to the pipe file descriptors */
	if (close(STDIN_FILENO) != 0)
		fprintf(stderr,"could not close stdin\n");
	if (dup(pipe2[0]) != STDIN_FILENO)
		fprintf(stderr,"could not duplicate file descriptor\n");
	/* reset stdout */
	if (close(STDOUT_FILENO) != 0)
		fprintf(stderr,"child: could not close stdout\n");
	if (dup(pipe1[1]) != STDOUT_FILENO)
		fprintf(stderr,"child: could not duplicate file descriptor\n");

        /* close all the pipes */
        if (close(pipe1[0]) != 0)
		fprintf(stderr,"child: could not close pipe1[0]\n");
        if (close(pipe1[1]) != 0)
		fprintf(stderr,"child: could not close pipe1[1]\n");
        if (close(pipe2[0]) != 0)
		fprintf(stderr,"child: could not close pipe2[0]\n");
        if (close(pipe2[1]) != 0)
		fprintf(stderr,"child: could not close pipe2[1]\n");

        /* start the command */
	exit(system(stream));
    }
}
return ;
}

/* build an input, output, or pipe object */
void build_input(root,parent,type)
     entry *root;
     Objdef *parent;
     char *type;
{
    Objdef *object;
    struct inputinfo *input_info;
    char defname[12];
    static int numinput=1;
    static int numoutput=1;
    static int numpipe=1;
    char *stream, *runflg;
    extern void input_update();
    int inpipe, outpipe;
    
    /* create new object */
    object = new_object(parent);

    input_info = (struct inputinfo*) XtMalloc( sizeof(struct inputinfo) );

    /* construct default input name */
    if (!strcmp(type,"input"))
       sprintf(defname,"input%d",numinput++);
    else if (!strcmp(type,"output"))
       sprintf(defname,"output%d",numoutput++);
    else 
       sprintf(defname,"pipe%d",numpipe++);
    

    /* find label, name, action in tree */
    object->name = get_value(root,"name",defname);
    input_info->type = XtNewString(type);
    input_info->separator = get_value(root,"separator","\n");
    object->action = parse_actions(object->name,root);
    object->info = input_info;
    object->sync = input_sync;
    if (!strcmp(type,"output") || !strcmp(type,"pipe")) 
	object->updater = input_update;

    /* get the stream attribute */
    stream =  get_value(root,"stream","none");

    /* allocate file descriptors */
    if (!strcmp(type,"input")) {
       input_info->fd = get_fd( stream, object->name, type );
       input_info->ofd = -1;
    } else if (!strcmp(type,"output")) {
       input_info->fd = -1;
       input_info->ofd = get_fd( stream, object->name, type );
    } else {
       get_pipes( stream, object->name, &inpipe, &outpipe );
       input_info->fd = inpipe;
       input_info->ofd = outpipe;
    }

    input_info->id = 0;

    /* get the running attribute */
    if( (!strcmp(type,"input")  && (input_info->fd == -1) ) ||
        (!strcmp(type,"output") && (input_info->ofd == -1) ) ||
        (!strcmp(type,"pipe")   && ((input_info->fd == -1 ) ||
				   (input_info->ofd == -1 )))) {
	input_info->running = 0;
    }else{
        runflg = get_value(root,"running","True");
        input_info->running = get_boolean(parent->widgetname,runflg);
    }

    /* add a callback for input available on this fd */
    if( input_info->running && (!strcmp(type,"input") ||
				!strcmp(type,"pipe"))) {
        input_info->id = XtAppAddInput( app_con, input_info->fd, 
	(XtPointer)XtInputReadMask, input_callback, (XtPointer)object);
    }
}

/* callback used for inputs and pipes */
void input_callback( client_data, source, id)
XtPointer client_data ;
int* source;
XtInputId* id;
{
    Objdef *object;
    struct inputinfo *input_info;

#define RBUFLEN 1024
    char readbuf[RBUFLEN];

    static string_buf *buffer=0;
    int nread,len;
    char *pos, *endpos, *next;

    if( buffer == 0 ) buffer = buf_start();
    
    object = (Objdef *) client_data;
    input_info = (struct inputinfo*)object->info;

    if( input_info->id != *id ) {
	fprintf(stderr,"ID mismatch in input_callback\n");
	exit(-1);
    }

/* 
 * We read to fill up readbuf (or however much is available )
 */

    nread = read( input_info->fd, readbuf, RBUFLEN-1 );

    if( nread < 0 ){
	perror("input_callback()");
	fprintf(stderr,"I/O error in input for object %s\n",object->name);
	exit(-1);
    }

    if( nread == 0 ){ /* EOF */
	/* no more input will be read */
	XtRemoveInput( input_info->id );
        input_info->id = 0;
	close(input_info->fd);
	input_info->fd = -1;
	input_info->running = 0;

	/* put a separator character in the readbuf */
	readbuf[0] = (input_info->separator)[0];
	nread = 1;
	/* then fall through so that anything remaining gets processed */
    }
    

    /* add the new input to the buffer */
    buf_cat( buffer, readbuf, nread );

    /* if there is one or more separators in the input we need to 
       process the buffer */
    readbuf[nread] = '\0';
    if( strpbrk( readbuf, input_info->separator  ) != 0 ){
	
	pos = buf_fetch(buffer); /* gets string and sets buffer to empty */
	endpos = pos + strlen(pos);

	/* find the next separator */
	next = strpbrk( pos, input_info->separator );

        while( next != 0 ){
	   
          if( (len = next - pos ) != 0 ){ 

	     /* copy the string to the object value and call the action */
             if( object->value != 0 ) XtFree( object->value );
	     object->value = (char*)XtMalloc( len +1 );
	     strncpy( object->value, pos, len );
	     object->value[len] = '\0';
             /* Do the actions[s] */
             perform_actions(object->name,object->action,1);

          }

	  /* find the next separator */
	  pos = next+1;
	  next = strpbrk( pos, input_info->separator );
        }

	/* anything left is put back at the start of the buffer */
	if( pos != endpos ){
	    buf_cat( buffer, pos, strlen(pos) );
	}

    }
}

/*
 * this routine is called when a input's state (running
 * or not) or stream is changed by the SET action.
 */
void input_sync(object,tag,value)
Objdef* object;
char* tag;
char* value;
{
    struct inputinfo *input_info;
    int inpipe, outpipe;

    input_info = (struct inputinfo *) object->info;

    /* remove old input callback so we can start over */
    if( input_info->running && (!strcmp(input_info->type,"input") ||
				!strcmp(input_info->type,"pipe"))) {
       if( input_info->id != 0 ) XtRemoveInput(input_info->id);
    }

    if (!strcmp(tag,"running")) {
       input_info->running = get_boolean(object->parent->widgetname,value);
    }

    if (!strcmp(tag,"stream")) {
       input_info->id = 0;
       /* close up the old fd */
       if( input_info->fd != -1 ) close(input_info->fd);
       input_info->fd = -1;
       if( input_info->ofd != -1 ) close(input_info->ofd);
       input_info->ofd = -1;
       /* open a new stream */
       if (!strcmp(input_info->type,"input")) {
         input_info->fd = get_fd( value, object->name, input_info->type );
       } else if (!strcmp(input_info->type,"output")) {
         input_info->ofd = get_fd( value, object->name, input_info->type );
       } else {
         get_pipes( value, object->name, &inpipe, &outpipe );
         input_info->fd = inpipe;
         input_info->ofd = outpipe;
       }
    }

    if( (!strcmp(input_info->type,"input")  && (input_info->fd == -1) ) ||
        (!strcmp(input_info->type,"output") && (input_info->ofd == -1) ) ||
        (!strcmp(input_info->type,"pipe")   && ((input_info->fd == -1 ) ||
				   (input_info->ofd == -1 )))) {
	input_info->running = 0;
    }

    if( input_info->running && (!strcmp(input_info->type,"input") ||
				!strcmp(input_info->type,"pipe"))) {
         input_info->id = XtAppAddInput( app_con, input_info->fd, 
		   (XtPointer)XtInputReadMask, input_callback, 
		                    (XtPointer)object);
    } 
}

void input_update(object,value)
Objdef* object;
char* value;
{
    struct inputinfo *input_info;
    int nwrite;

    input_info = (struct inputinfo *) object->info;

    /* only a pipe or output object can write to the process */
    if (!strcmp(input_info->type,"input")) return;

    nwrite = write( input_info->ofd, value, strlen(value) );

    if( nwrite < strlen(value) ){
	perror("input_update()");
	fprintf(stderr,"I/O error trying to update object %s\n",object->name);
	exit(-1);
    }
}
